/* ========================================================================= */
/*   Main.c                                                                  */
/*   (c) 2007 St                                                             */
/*                                                                           */
/*   Program for nixie clock controlled MC68HC908QY1                         */
/*   Xtall frequency: 12 MHz                                                 */
/* ========================================================================= */
#include <hidef.h> /* for EnableInterrupts macro */
#include "derivative.h" /* include peripheral declarations */

// Definition of PIN connection
#define POSITION0 PTA_PTA3  // digits anode switches
#define POSITION1 PTB_PTB4
#define POSITION2 PTB_PTB5
#define POSITION3 PTB_PTB6
#define POSITION4 PTB_PTB7
#define POSITION5 PTB_PTB1

#define DIGIT0 PTA_PTA0     // output multiplex
#define DIGIT1 PTB_PTB3 
#define DIGIT2 PTB_PTB2 
#define DIGIT3 PTA_PTA1

#define BUZZER PTB_PTB0     // alarm buzzer
#define KEYINP PTA_PTA2     // button input

// Connection of function keys to digit positions
#define KEYMENU  4 
#define KEYUP    3 
#define KEYENTER 2 

// Global variables used in main loop and in interrupt
// counter for miliseconds
word milisec = 0;
// arrays for real time, alarm and auxiliary time during setting
byte time[6],alarm[6],auxtime[6];
// pressed code button, half second flag (for alarm and blink), blinking position 
byte button = 0, halfsec = 0, blink = 0xF;
// pointer which array to display
byte *p_digit = time;

// wait for button pressed or 20 second time-out
void waitbutton(void) {
  word i;
  byte n;
  
  for(n=0; n<200; n++) {
    // Wait for about 100 msec
    for(i=0; i<10000; i++) {asm sta COPCTL;} /* feeds the dog */
    if(button) break;
  }
}

// wait for all buttons released
void endbutton(void) {
  word i;

  while(button) {
    button = 0;
    for(i=0; i<10000; i++) {asm sta COPCTL;} /* feeds the dog */
  }
}

// function for setting array "setdata" with a new time
// seconds are not updated, just only hours and minutes
byte settime(byte *setdata) {
  byte posit = 2, updated = 0, next = 1;

  while(next) {
    blink = posit;   // updated position blinks
    waitbutton();
    // Test buttons and do 1 of 4 possibilities
    switch(button) {
      case KEYMENU : next = 0; break; // Stop updating time
      case KEYENTER: {  // Move to next digit
        posit++;
        if(posit == 6) posit = 2;
      } break;
      case KEYUP   : {  // Increment digit on this position
        setdata[posit]++;
        if(((posit == 2) || (posit == 4)) && (setdata[posit] == 10)) setdata[posit]=0;
        if((posit == 3) && (setdata[posit] == 6)) setdata[posit]=0;
        if((posit == 5) && (setdata[posit] == 3)) setdata[posit]=0;
        updated = 1;
      } break;
      default      : next = 0;  // Time-out expired
    }
    endbutton();
  }
  // Hours above 23 are not allowed
  if((setdata[5] == 2) && (setdata[4] > 3)) {
    setdata[5] = 0; setdata[4] = 0;
  }
  blink = 0xF;
  return(updated);
}
// ****************************
//    Main program function  //
// ****************************
void main(void) {
  byte i; // auxiliary counter for delay loops

  EnableInterrupts; /* enable interrupts */
  /* include your code here */

  // set direction of ports
  DDRA = 0x1B; DDRB = 0xFF;
  
  // ***  Internal to external oscillator switching START ***
  // according data sheet - paragraph 8.4.1.2
  PTA_PTA4 = 1;       // charge Xtal
  for(i=1; i<5; i++) {asm NOP;} // delay for charge
  // Switch OFF LVI detector (voltage is lower than 5 V)
  CONFIG1 = 0x30;
  // External XTAL oscillator, pullup disconnected 
  CONFIG2 = 0x98;
  for(i=1; i<255; i++) {asm sta COPCTL;} // delay for start oscillation
  OSCSTAT = 0x2;
  // ***  Internal to external oscillator switching END ***

  // Default values for start
  time[4] = 7; time[3] = 3;
  alarm[4] = 7; alarm[3] = 4;
  alarm[1] = 0xF; alarm[0] = 0xF;

  /* Init_TIM (timer) for 1 milisec period */
  /* TSC: TOF=0,TOIE=0,TSTOP=1,TRST=1,PS2=0,PS1=0,PS0=0 */
  TSC = 0x30;                          /* Stop and reset counter */
  TMOD = 0x0BB7;                       /* Period value setting */
  (void)(TSC == 0);                    /* Overflow int. flag clearing (first part) */
  /* TSC: TOF=0,TOIE=1,TSTOP=1,TRST=0,PS2=0,PS1=0,PS0=0 */
  TSC = 0x40;                          /* Int. flag clearing (2nd part) and timer contr. register setting */

  /***** Main loop of program *****/
  for(;;) {
    {asm sta COPCTL;} /* feeds the dog */
    // Set alarm Buzzer if time == alarm time
    if((time[5] == alarm[5]) && (time[4] == alarm[4]) && (time[3] == alarm[3]) && (time[2] == alarm[2]) && halfsec) BUZZER = 0; else BUZZER = 1;
    
    if(button == KEYMENU) {
    // Menu button enters time updating
      endbutton();
      p_digit = alarm;
      if(!settime(alarm)) {
        // Alarm was not updated, try Time
        DisableInterrupts;
        // We must not destroy real Time, copy it to auxiliary array
        auxtime[5] = time[5];auxtime[4] = time[4];auxtime[3] = time[3];auxtime[2] = time[2];
        EnableInterrupts;
        p_digit = auxtime;
        if(settime(auxtime)) {
          // Time was updated - copy it back
          DisableInterrupts;
          time[5] = auxtime[5];time[4] = auxtime[4];time[3] = auxtime[3];
          time[2] = auxtime[2];time[1] = 0;time[0] = 0;
          EnableInterrupts;
        }
      }

      p_digit = time;
    } else button = 0;
      
  } /* loop forever */
  /* please make sure that you never leave main */
}

/*
** ===================================================================
**     Interrupt handler : isrINT_TIMOvr
** ===================================================================
*/
__interrupt void isrINT_TIMOvr(void)
{
  static const byte tabledigit[] = {3,2,7,4,1,9,8,0,5,6,0xF,0xF,0xF,0xF,0xF,0xF};
  static byte position = 0;
  static byte auxdigit;

  (void)(TSC == 0);         /* Overflow int. flag clearing (first part) */
  TSC = 0x40;               /* clear TOF */
  
  // Increment Time
  milisec++;
  if(milisec == 500) halfsec = 1;
  if(milisec == 1000) {
    halfsec = 0;                 // half second flag
    milisec = 0;                 // miliseconds
    time[0]++;
    if(time[0] == 10) {          // seconds
      time[0] = 0;
      time[1]++;
      if(time[1] == 6) {
        time[1] = 0;
        time[2]++;
        if(time[2] == 10) {      // minutes
          time[2] = 0;
          time[3]++;
          if(time[3] == 6) {
            time[3] = 0;
            time[4]++;
            if(time[4] == 10) {  // hours
              time[4] = 0;
              time[5]++;
            }
            if((time[4] == 4) && (time[5] == 2)) {
              time[4] = 0;
              time[5] = 0;
            }
          }
        }
      }
    }
  }
  // Clear display (all digits off)
  POSITION0 = 0; POSITION1 = 0; POSITION2 = 0; POSITION3 = 0; POSITION4 = 0; POSITION5 = 0;
      
  // New digit to display
  // Test for blinking digit (and set it according "blink" variable)
  if((blink == position) && halfsec) {
    DIGIT0 = 1; DIGIT1 = 1; DIGIT2 = 1; DIGIT3 = 1;
  } else {
    // Set new digit to display
    auxdigit = tabledigit[*(p_digit + position)];
    DIGIT0 = 0; DIGIT1 = 0; DIGIT2 = 0; DIGIT3 = 0;
    if(auxdigit & 0x1) DIGIT0 = 1;
    if(auxdigit & 0x2) DIGIT1 = 1;
    if(auxdigit & 0x4) DIGIT2 = 1;
    if(auxdigit & 0x8) DIGIT3 = 1;
  }

  // Switch ON new digit
  switch(position) {
    case 0: POSITION0 = 1; break;
    case 1: POSITION1 = 1; break;
    case 2: POSITION2 = 1; break;
    case 3: POSITION3 = 1; break;
    case 4: POSITION4 = 1; break;
    case 5: POSITION5 = 1; break;
  }

  // Test if pressed button with this position  
  if((!button) && (PTA & 0x4)) button = position;
  
  // Update position for next digit
  position++;
  if(position == 6) position = 0;
}
/* end of isrINT_TIMOvr */

/* Initialization of the CPU registers in FLASH */
#define UNASSIGNED_ISR 0xFFFF          /* Unassigned interrupt service routine */

extern void _Startup(void);            /* reset interrupt service routine */

void (* const _vect[])() @0xFFE0 = {   // Interrupt vector table
        UNASSIGNED_ISR,                /* Int.no.  0 INT_KBI (at FFE0)               Unassigned */
        UNASSIGNED_ISR,                /* Int.no.  1 Reserved1 (at FFE2)             Unassigned */
        UNASSIGNED_ISR,                /* Int.no.  2 Reserved2 (at FFE4)             Unassigned */
        UNASSIGNED_ISR,                /* Int.no.  3 Reserved3 (at FFE6)             Unassigned */
        UNASSIGNED_ISR,                /* Int.no.  4 Reserved4 (at FFE8)             Unassigned */
        UNASSIGNED_ISR,                /* Int.no.  5 Reserved5 (at FFEA)             Unassigned */
        UNASSIGNED_ISR,                /* Int.no.  6 Reserved6 (at FFEC)             Unassigned */
        UNASSIGNED_ISR,                /* Int.no.  7 Reserved7 (at FFEE)             Unassigned */
        UNASSIGNED_ISR,                /* Int.no.  8 Reserved8 (at FFF0)             Unassigned */
        isrINT_TIMOvr,                 /* Int.no.  9 INT_TIMOvr (at FFF2)            Used */
        UNASSIGNED_ISR,                /* Int.no. 10 INT_TIMCH1 (at FFF4)            Unassigned */
        UNASSIGNED_ISR,                /* Int.no. 11 INT_TIMCH0 (at FFF6)            Unassigned */
        UNASSIGNED_ISR,                /* Int.no. 12 Reserved12 (at FFF8)            Unassigned */
        UNASSIGNED_ISR,                /* Int.no. 13 INT_IRQ (at FFFA)               Unassigned */
        UNASSIGNED_ISR,                /* Int.no. 14 INT_SWI (at FFFC)               Unassigned */
};

/* END */

